;; Generated package description from clipboard-collector.el  -*- no-byte-compile: t -*-
(define-package "clipboard-collector" "0.3" "Collect clipboard entries according to regex rules" '((emacs "25")) :commit "ee4a0ee506c47666714cd61649334936d67e5a43" :authors '(("Clemens Radermacher" . "clemera@posteo.net")) :maintainer '("Clemens Radermacher" . "clemera@posteo.net") :keywords '("convenience") :url "https://github.com/clemera/clipboard-collector")
